from typing import List, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    from .view import View


class ObservableObject:
    def __init__(self) -> None:
        self._watch_variables: List[str] = []
        self.view: Optional['View'] = None

    def notify(self) -> None:
        """
        update view
        :return: None
        """
        if self.view is not None:
            self.view.update_sync()



from __future__ import annotations
from typing import Optional, Union, Callable

import discord
from discord import ui
from discord.utils import MISSING

from .utils import _call_any
from .modal import Modal


class CustomButton(ui.Button):
    def __init__(
            self,
            label: str = "",
            style: discord.ButtonStyle = discord.ButtonStyle.primary,
            disabled: bool = False,
            emoji: Optional[Union[str, discord.PartialEmoji]] = None,
            custom_id: Optional[str] = None,
            modal_submit: Optional[Modal] = None
    ):
        super().__init__(label=label, style=style, disabled=disabled, emoji=emoji, custom_id=custom_id)
        self.callback_func: Optional[Callable] = None
        self.check_func: Optional[Callable[[discord.Interaction], bool]] = None
        self.modal_submit = modal_submit

    async def callback(self, interaction: discord.Interaction) -> None:
        if self.callback_func is None and self.modal_submit is None:
            return
        if self.check_func is not None:
            if not self.check_func(interaction):
                return
        if self.modal_submit is not None:
            await interaction.response.send_modal(self.modal_submit)
            return
        await _call_any(self.callback_func, interaction)


class CustomSelect(ui.Select):
    def __init__(
            self,
            *,
            custom_id: Optional[str],
            placeholder: Optional[str] = None,
            min_values: int = 1,
            max_values: int = 1,
            options: Optional[list[discord.SelectOption]],
            row: Optional[int] = None,
            callback: Optional[Callable] = None,
            check_func: Callable[[discord.Interaction], bool]
    ) -> None:
        custom_id = custom_id or MISSING
        options = options or MISSING
        super(CustomSelect, self).__init__(
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            options=options,
            row=row
        )
        self.callback_func = callback
        self.check_func = check_func

    async def callback(self, interaction: discord.Interaction) -> None:
        if self.callback_func is None:
            return
        if self.check_func is not None:
            if not self.check_func(interaction):
                return
        selected_options = []
        for label in interaction.data.get("values", []):
            for option in self.options:
                if option.label == label:
                    selected_options.append(option)
                    continue
        await _call_any(self.callback_func, interaction, selected_options)



from __future__ import annotations
from typing import Union

import discord
from discord import ui

from .item import Item


class Message:
    def __init__(
            self,
            content: str = "",
            embeds: list[discord.Embed] = None,
            components: list[Union[list[Item], Item]] = None):
        self._content = content
        self._embeds: list[discord.Embed] = embeds or []
        self._components: list[Union[list[Item], Item]] = components or []

    def content(self, content: str) -> Message:
        self._content = content
        return self

    def embed(self, embed: discord.Embed) -> Message:
        self._embeds.append(embed)
        return self

    def embeds(self, embeds: list[discord.Embed]) -> Message:
        self._embeds.extend(embeds)
        return self

    def item(self, item: Union[list[Item], Item]) -> Message:
        self._components.append(item)
        return self

    def items(self, items: list[Union[list[Item], Item]]) -> Message:
        self._components.extend(items)
        return self

    def get_discord_items(self) -> list[ui.Item]:
        row = 0
        items = []
        for component in self._components:
            if isinstance(component, list):
                for sub_component in component:  # type: Item
                    items.append(sub_component.to_discord_item(row))
                row += 1
            else:
                items.append(component.to_discord_item(None))

        return items

    def __eq__(self, other: Message) -> bool:
        return self._components == other._components\
               and self._embeds == other._embeds\
               and self._content == other._content



from __future__ import annotations

import asyncio
from typing import Any, Callable, Optional

from .publisher import Publisher, MapPublisher


async def _call_any(func: Callable, *args: Any, **kwargs: Any) -> Any:
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    return func(*args, **kwargs)


class AsyncPublisher(Publisher):
    def __init__(self):
        super().__init__()
        self.child: Optional[AsyncPublisher] = None
        self.parent: Optional[AsyncPublisher] = None
        self.subscribers: list[Callable[[Any], None]] = []

    def chain(self, new_publisher: AsyncPublisher) -> AsyncPublisher:
        new_publisher.parent = self
        self.set_child(new_publisher)
        return new_publisher

    async def downstream(self, value: Any):
        if self.parent is not None:
            await self.parent.downstream(value)

    async def upstream(self, value: Any):
        if self.child is not None:
            await self.child.upstream(value)

        for func in self.subscribers:
            await _call_any(func, value)

    async def dispatch(self) -> Any:
        if self.parent is not None:
            await self.parent.dispatch()

    async def sink(self, func: Callable[[Any], None]) -> AsyncPublisher:
        if self.child is not None:
            await self.child.sink(func)
            return self
        self.subscribers.append(func)
        await self.dispatch()
        return self

    def map(self, func: Callable[[Any], Any]) -> AsyncPublisher:
        if self.child is not None:
            self.child.map(func)
            return self
        new_publisher = AsyncMapPublisher(func)
        self.chain(new_publisher)
        return self


class AsyncMapPublisher(MapPublisher, AsyncPublisher):
    def __init__(self, func: Callable[[Any], Any]) -> None:
        super().__init__(func)
        self.map_func = func

    async def upstream(self, value: Any):
        if isinstance(value, list):
            value = [await _call_any(self.map_func, v) for v in value]
        else:
            value = await _call_any(self.map_func, value)

        for func in self.subscribers:
            await _call_any(func, value)



from __future__ import annotations
from typing import Any

from .publisher import Publisher


class Just(Publisher):
    def __init__(self, value: Any) -> None:
        super().__init__()
        self.value: Any = value

    def dispatch(self) -> Any:
        self.upstream(self.value)

    def send(self, value: Any) -> None:
        raise NotImplementedError



from __future__ import annotations

import asyncio
from typing import Any

import discord

from .view import View
from .button import Button
from .message import Message
from .tracker import ViewTracker
from .provider import InteractionProvider


class Alert(View):
    def __init__(self, title: str, text: str, buttons: list[ActionButton | list[ActionButton]], ephemeral: bool = False):
        super().__init__()
        self.title = title
        self.text = text
        self.buttons = buttons
        self.ephemeral = ephemeral
        self.clicked = asyncio.Future()

    def _set_func(self):
        buttons = []
        for button in self.buttons:
            if isinstance(button, list):
                buttons_ = []
                for button_ in button:  # type: ActionButton
                    buttons_.append(button_.set_future(self.clicked))
                buttons.append(buttons_)
            else:
                buttons.append(button.set_future(self.clicked))

        self.buttons = buttons

    async def body(self) -> Message | View:
        self._set_func()
        return Message(embeds=[
            discord.Embed(title=self.title, description=self.text, colour=discord.Colour.blurple())
        ], components=self.buttons)

    async def wait_for_click(self, interaction: discord.Interaction, timeout: float | None = 180.0):
        tracker = ViewTracker(self, timeout)
        await tracker.track(InteractionProvider(interaction, ephemeral=self.ephemeral))
        await self.clicked
        self.stop()
        return self.clicked.result()


class ActionButton(Button):
    def __init__(self, label: str, style: discord.ButtonStyle, value: Any, **kwargs):
        super().__init__(label=label, style=style, **kwargs)
        self.value = value
        self.clicked: asyncio.Future | None = None

    def set_future(self, future: asyncio.Future) -> Button:
        self.clicked = future
        super(ActionButton, self).on_click(self.selected)
        return self

    def selected(self, _: discord.Interaction):
        self.clicked.set_result(self.value)
